"""
Utils functions of demeter
"""

from .application import float_param_formatter, to_decimal
from .console_text import (
    ForColorEnum,
    BackColorEnum,
    ModeEnum,
    get_formatted,
    STYLE,
    get_formatted_predefined,
    get_formatted_from_dict,
)
